# coding=utf-8
"""
Contains the events manager module for handling event related functionality of data fetching.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from typing import Dict, Any, List, Type

from web3.contract import Contract

from fastlane_bot.events.managers.base import BaseManager


class EventManager(BaseManager):
    @property
    def events(self) -> List[Type[Contract]]:
        """
        Get the events from the exchanges.

        Returns
        -------
        List[Type[Contract]]
            The events.
        """
        return [
            event
            for exchange in self.exchanges.values()
            for event in exchange.get_events(
                self.event_contracts[exchange.exchange_name]
            )
        ]

    def add_pool_info_from_event(
        self, strategy: List[Any], block_number: int = None
    ) -> Dict[str, Any]:
        """
        Add the pool info from the strategy.

        Parameters
        ----------
        strategy : List[Any]
            The strategy.
        block_number (optional) : int
            The block number the data was gathered.

        Returns
        -------
        Dict[str, Any]
            The pool info.

        """
        cid = strategy[0]
        order0, order1 = strategy[3][0], strategy[3][1]
        tkn0_address, tkn1_address = self.web3.toChecksumAddress(
            strategy[2][0]
        ), self.web3.toChecksumAddress(strategy[2][1])

        return self.add_pool_info(
            address=self.cfg.CARBON_CONTROLLER_ADDRESS,
            exchange_name="carbon_v1",
            fee="0.002",
            fee_float=0.002,
            tkn0_address=tkn0_address,
            tkn1_address=tkn1_address,
            cid=cid,
            other_args=dict(
                y_0=order0[0],
                y_1=order1[0],
                z_0=order0[1],
                z_1=order1[1],
                A_0=order0[2],
                A_1=order1[2],
                B_0=order0[3],
                B_1=order1[3],
            ),
            block_number=block_number,
        )



# coding=utf-8
"""
Contains the pool class for Carbon v1. This class is responsible for handling Carbon v1 pools and updating the state of the pools.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from dataclasses import dataclass
from typing import Dict, Any, Tuple, List

from web3.contract import Contract

from .base import Pool


@dataclass
class CarbonV1Pool(Pool):
    """
    Class representing a Carbon v1 pool.
    """

    exchange_name: str = "carbon_v1"

    @staticmethod
    def unique_key() -> str:
        """
        see base class.
        """
        return "cid"

    @classmethod
    def event_matches_format(cls, event: Dict[str, Any]) -> bool:
        """
        see base class.
        """
        event_args = event["args"]
        return "order0" in event_args

    def update_from_event(
        self, event_args: Dict[str, Any], data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        See base class.
        """
        event_type = event_args["event"]
        data = CarbonV1Pool.parse_event(data, event_args, event_type)
        for key, value in data.items():
            self.state[key] = value
        return data

    @staticmethod
    def parse_event(
        data: Dict[str, Any], event_args: Dict[str, Any], event_type: str
    ) -> Dict[str, Any]:
        """
        Parse the event args into a dict.

        Parameters
        ----------
        data : Dict[str, Any]
            The data to update.
        event_args : Dict[str, Any]
            The event arguments.
        event_type : str
            The event type.

        Returns
        -------
        Dict[str, Any]
            The updated data.
        """
        order0, order1 = CarbonV1Pool.parse_orders(event_args, event_type)
        data["cid"] = event_args["args"].get("id")
        data["y_0"] = order0[0]
        data["z_0"] = order0[1]
        data["A_0"] = order0[2]
        data["B_0"] = order0[3]
        data["y_1"] = order1[0]
        data["z_1"] = order1[1]
        data["A_1"] = order1[2]
        data["B_1"] = order1[3]
        return data

    @staticmethod
    def parse_orders(
        event_args: Dict[str, Any], event_type: str
    ) -> Tuple[List[int], List[int]]:
        """
        Parse the orders from the event args. If the event type is StrategyDeleted, then the orders are set to 0.

        Parameters
        ----------
        event_args : Dict[str, Any]
            The event arguments.
        event_type : str
            The event type.

        Returns
        -------
        Tuple[List[int], List[int]]
            The parsed orders.
        """
        if event_type != "StrategyDeleted":
            order0 = event_args["args"].get("order0")
            order1 = event_args["args"].get("order1")
        else:
            order0 = [0, 0, 0, 0]
            order1 = [0, 0, 0, 0]
        return order0, order1

    def update_from_contract(self, contract: Contract) -> Dict[str, Any]:
        """
        See base class.
        """
        try:
            strategy = contract.strategy(self.state["cid"])
        except AttributeError:
            strategy = contract.caller.strategy(self.state["cid"])

        fake_event = {
            "args": {
                "id": strategy[0],
                "order0": strategy[3][0],
                "order1": strategy[3][1],
            }
        }
        params = self.parse_event(self.state, fake_event, "None")
        params["fee"] = "0.002"
        params["fee_float"] = 0.002
        params["exchange_name"] = "carbon_v1"
        for key, value in params.items():
            self.state[key] = value
        return params



"""
Fastlane bot config -- database configuration
"""
__VERSION__ = "1.0.4"
__DATE__ = "01/May 2023"
from .base import ConfigBase
from . import selectors as S

import os
from dotenv import load_dotenv
load_dotenv()



class ConfigDB(ConfigBase):
    """
    Fastlane bot config -- database
    """
    __VERSION__=__VERSION__
    __DATE__=__DATE__
    
    _PROJECT_PATH = os.path.normpath(f"{os.getcwd()}") # TODO: FIX THIS
    
    DATABASE_SEED_FILE = os.path.normpath(
        f"{_PROJECT_PATH}/fastlane_bot/data/seed_token_pairs.csv"
    )
    
    
    DATABASE_SQLITE = S.DATABASE_SQLITE
    DATABASE_POSTGRES = S.DATABASE_POSTGRES
    DATABASE_MEMORY = S.DATABASE_MEMORY
    DATABASE_SDK = S.DATABASE_SDK
    DATABASE_UNITTEST = S.DATABASE_UNITTEST
    @classmethod
    def new(cls, db=None, **kwargs):
        """
        Return a new ConfigDB object for the specified database.
        """
        if db is None:
            db = cls.DATABASE_POSTGRES
        if db == cls.DATABASE_POSTGRES:
            return _ConfigDBPostgres(_direct=False, **kwargs)
        elif db == cls.DATABASE_SQLITE:
            return _ConfigDBSqlite(_direct=False, **kwargs)
        elif db == cls.DATABASE_MEMORY:
            return _ConfigDBMemory(_direct=False, **kwargs)
        elif db == cls.DATABASE_SDK:
            return _ConfigDBSdk(_direct=False, **kwargs)
        elif db == cls.DATABASE_UNITTEST:
            return _ConfigDBUnitTest(_direct=False, **kwargs)
        else:
            raise ValueError(f"Invalid db: {db}")
        
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

class _ConfigDBPostgres(ConfigDB):
    """
    Fastlane bot config -- database [Postgres]
    """
    DATABASE = S.DATABASE_POSTGRES    
    POSTGRES_USER = os.environ.get("POSTGRES_USER")
    POSTGRES_PASSWORD = os.environ.get("POSTGRES_PASSWORD")
    POSTGRES_HOST = os.environ.get("POSTGRES_HOST")
    POSTGRES_DB = os.environ.get("POSTGRES_DB")
    POSTGRES_URL = None # set in init
    
    POSTGRES_USER_DEFAULT = "postgres"
    POSTGRES_PASSWORD_DEFAULT = "postgres"
    POSTGRES_HOST_DEFAULT = "localhost"
    POSTGRES_DATABASE_DEFAULT = "mainnet"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.POSTGRES_URL is None:
            # those are all set to None if URL is given
            self.POSTGRES_USER = None
            self.POSTGRES_PASSWORD = None
            self.POSTGRES_HOST = None
            self.POSTGRES_DB = None
            return
            
        if not self.POSTGRES_USER:
            self.POSTGRES_USER = self.POSTGRES_USER_DEFAULT
        if not self.POSTGRES_PASSWORD:
            self.POSTGRES_PASSWORD = self.POSTGRES_PASSWORD_DEFAULT
        if not self.POSTGRES_HOST:
            self.POSTGRES_HOST = self.POSTGRES_HOST_DEFAULT
        if not self.POSTGRES_DB:
            self.POSTGRES_DB = self.POSTGRES_DATABASE_DEFAULT
        self.POSTGRES_URL = f"postgresql://{self.POSTGRES_USER}:{self.POSTGRES_PASSWORD}@{self.POSTGRES_HOST}/{self.POSTGRES_DB}"
        self.DEFAULT_DB_BACKEND_URL = self.POSTGRES_URL
            
class _ConfigDBSqlite(ConfigDB):
    """
    Fastlane bot config -- database [Sqlite]
    """
    DATABASE = S.DATABASE_SQLITE
    SQLITE_URL = "sqlite:///fastlane_bot.db"
    #DEFAULT_DB_BACKEND_URL = POSTGRES_URL
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        raise NotImplementedError("Sqlite not implemented")

class _ConfigDBUnittest(ConfigDB):
    """
    Fastlane bot config -- database [Postgres]
    """
    DATABASE = S.DATABASE_UNITTEST

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    
class _ConfigDBMemory(ConfigDB):
    """
    Fastlane bot config -- database [Memory]
    """
    DATABASE = S.DATABASE_MEMORY
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        raise NotImplementedError("Memory not implemented")

    
class _ConfigDBSdk(ConfigDB):
    """
    Fastlane bot config -- database [SDK]
    """
    DATABASE = S.DATABASE_SDK
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        raise NotImplementedError("SDK not implemented")

class _ConfigDBUnitTest(ConfigDB):
    """
    Fastlane bot config -- database [UnitTest]
    """
    DATABASE = S.DATABASE_UNITTEST
    POSTGRES_USER = os.environ.get("POSTGRES_USER")
    POSTGRES_PASSWORD = os.environ.get("POSTGRES_PASSWORD")
    POSTGRES_HOST = os.environ.get("POSTGRES_HOST")
    POSTGRES_DB = os.environ.get("POSTGRES_DB")
    POSTGRES_URL = None # set in init

    POSTGRES_USER_DEFAULT = "postgres"
    POSTGRES_PASSWORD_DEFAULT = "postgres"
    POSTGRES_HOST_DEFAULT = "localhost"
    POSTGRES_DATABASE_DEFAULT = "unittest"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if not self.POSTGRES_URL is None:
            # those are all set to None if URL is given
            self.POSTGRES_USER = None
            self.POSTGRES_PASSWORD = None
            self.POSTGRES_HOST = None
            self.POSTGRES_DB = None
            return

        if not self.POSTGRES_USER:
            self.POSTGRES_USER = self.POSTGRES_USER_DEFAULT
        if not self.POSTGRES_PASSWORD:
            self.POSTGRES_PASSWORD = self.POSTGRES_PASSWORD_DEFAULT
        if not self.POSTGRES_HOST:
            self.POSTGRES_HOST = self.POSTGRES_HOST_DEFAULT
        if not self.POSTGRES_DB:
            self.POSTGRES_DB = self.POSTGRES_DATABASE_DEFAULT
        self.POSTGRES_URL = f"postgresql://{self.POSTGRES_USER}:{self.POSTGRES_PASSWORD}@{self.POSTGRES_HOST}/{self.POSTGRES_DB}"
        self.DEFAULT_DB_BACKEND_URL = self.POSTGRES_URL
    


# coding=utf-8
"""
Contains the base class for exchanges. This class is responsible for handling exchanges and updating the state of the pools.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, List, Type, Any

from web3.contract import Contract

from fastlane_bot.events.pools.base import Pool


@dataclass
class Exchange(ABC):
    """
    Base class for exchanges
    """

    __VERSION__ = "0.0.1"
    __DATE__ = "2023-07-03"

    pools: Dict[str, Pool] = field(default_factory=dict)

    def get_pools(self) -> List[Pool]:
        """
        Get the pools of the exchange.

        Returns
        -------
        List[Pool]
            The pools of the exchange

        """
        return list(self.pools.values())

    @abstractmethod
    def add_pool(self, pool: Pool):
        """
        Add a pool to the exchange.

        Parameters
        ----------
        pool : Pool
            The pool object

        Returns
        -------
        None
        """
        pass

    @abstractmethod
    def get_abi(self):
        """
        Get the ABI of the exchange

        Returns
        -------
        ABI
            The ABI of the exchange

        """
        pass

    @abstractmethod
    def get_events(self, contract: Contract) -> List[Type[Contract]]:
        """
        Get the events of the exchange

        Parameters
        ----------
        contract : Contract
            The contract object

        Returns
        -------
        List[Type[Contract]]
            The events of the exchange

        """
        pass

    @abstractmethod
    def get_fee(self, address: str, contract: Contract) -> float:
        """
        Get the fee of the exchange

        Parameters
        ----------
        address : str
            The address of the exchange
        contract : Contract
            The contract object

        Returns
        -------
        float
            The fee of the exchange

        """
        pass

    @abstractmethod
    def get_tkn0(self, address: str, contract: Contract, event: Any) -> str:
        """
        Get the tkn0 of the exchange

        Parameters
        ----------
        address : str
            The address of the exchange
        contract : Contract
            The contract object
        event : Any
            The event object

        Returns
        -------
        str
            The tkn0 of the exchange
        """
        pass

    @abstractmethod
    def get_tkn1(self, address: str, contract: Contract, event: Any) -> str:
        """
        Get the tkn1 of the exchange

        Parameters
        ----------
        address : str
            The address of the exchange
        contract : Contract
            The contract object
        event : Any
            The event object

        Returns
        -------
        str
            The tkn1 of the exchange

        """
        pass

    def get_pool(self, key: str) -> Pool:
        """

        Parameters
        ----------
        key: str
            pool_address if UniswapV2 or SushiswapV2 or UniswapV3
            else tkn1_address if BancorV3
            else cid if Carbon

        Returns
        -------
        Pool
            The pool object

        """
        return self.pools[key] if key in self.pools else None



"""
Helpers for the Fastlane project.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
__VERSION__ = "1.2"
__DATE__="02/May/2023"

from dataclasses import dataclass
from typing import Union, Any
from _decimal import Decimal
from fastlane_bot.events.interface import Token, Pool


@dataclass
class TradeInstruction:
    """
    A class that handles the conversion of token decimals for the bot.

    Parameters
    ----------
    cid: str
        The pool unique ID
    tknin: str
        The input token key (e.g. 'DAI-1d46')
    amtin: int or Decimal or float
        The input amount.
    tknout: str
        The output token key (e.g. 'DAI-1d46')
    amtout: int or Decimal or float
        The output amount.
    cid_tkn: str
        If the curve is a Carbon curve, the cid will have a "-1" or "-0" to denote which side of the strategy the trade is on.
        This parameter is used to remove the "-1" or "-0" from the cid.
    raw_txs: str
    pair_sorting: str

    Attributes
    ----------
    
    TODO -- DOCSTRING OUT OF DATE 
    _tknin_address: str
        The input token address.
    _tknin_decimals: int
        The input token decimals.
    _amtin_wei: int
        The input amount in wei.
    _amtin_decimals: Decimal
        The input amount in decimals.
    _tknout_address: str
        The output token address.
    _tknout_decimals: int
        The output token decimals.
    _amtout_wei: int
        The output amount in wei.
    _amtout_decimals: Decimal
        The output amount in decimals.
    _is_carbon: bool
        Whether the curve is a Carbon curve.

    """
    __VERSION__=__VERSION__
    __DATE__=__DATE__
    
    ConfigObj: Any
    cid: str
    tknin: str
    amtin: Union[int, Decimal, float]
    tknout: str
    amtout: Union[int, Decimal, float]
    pair_sorting: str = None
    raw_txs: str = None
    custom_data: str = ''
    db: any = None
    tknin_dec_override: int = None   # for testing to not go to the database
    tknout_dec_override: int = None  # ditto
    tknin_addr_override: str = None  # ditto
    tknout_addr_override: str = None # ditto
    exchange_override: str = None    # ditto
    _amtin_wei: int = None
    _amtout_wei: int = None

    @property
    def tknin_key(self) -> str:
        """
        The input token key (e.g. 'DAI-1d46')
        """
        return self.tknin

    @property
    def tknout_key(self) -> str:
        """
        The output token key (e.g. 'DAI-1d46')
        """
        return self.tknout

    def __post_init__(self):
        """
        Use the database session to get the token addresses and decimals based on the Pool.cid and Token.key
        """
        self._cid_tkn: str = None
        self._is_carbon = self._check_if_carbon()
        
        if self.tknin_dec_override is None:
            TokenIn = self.db.get_token(key =self.tknin)
            self._tknin_address = TokenIn.address
            self._tknin_decimals = int(TokenIn.decimals)
        else:
            self._tknin_address = self.tknin_addr_override
            self._tknin_decimals = self.tknin_dec_override

        if self._amtin_wei is None:
            self._amtin_wei = self._convert_to_wei(self.amtin, self._tknin_decimals)

        self._amtin_decimals = self._convert_to_decimals(
            self.amtin, self._tknin_decimals
        )
        self._amtin_quantized = self._quantize(
            self._amtin_decimals, self._tknin_decimals
        )
        
        if self.tknout_dec_override is None:
            TokenOut = self.db.get_token(key =self.tknout)
            self._tknout_address = TokenOut.address
            self._tknout_decimals = int(TokenOut.decimals)
        else:
            self._tknout_address = self.tknout_addr_override
            self._tknout_decimals = self.tknout_dec_override
            
        if self._amtout_wei is None:            
            self._amtout_wei = self._convert_to_wei(self.amtout, self._tknout_decimals)

        self._amtout_decimals = self._convert_to_decimals(
            self.amtout, self._tknout_decimals
        )
        self._amtout_quantized = self._quantize(
            self._amtout_decimals, self._tknout_decimals
        )
        if self.raw_txs is None:
            self.raw_txs = "[]"
        if self.pair_sorting is None:
            self.pair_sorting = ""
        if self.exchange_override is None:
            self._exchange_name = self.db.get_pool(cid=self.cid.split('-')[0]).exchange_name
        else:
            self._exchange_name = self.exchange_override
        self._exchange_id = self.ConfigObj.EXCHANGE_IDS[self._exchange_name]

    @property
    def platform_id(self) -> int:
        """
        The exchange ID. (0 = Bancor V2, 1 = Bancor V3, 2 = Uniswap V2, 3 = Uniswap V3, 4 = Sushiswap V2, 5 = Sushiswap, 6 = Carbon)
        """
        return self._exchange_id

    @property
    def exchange_name(self) -> str:
        """
        The exchange name.
        """
        return self._exchange_name

    @staticmethod
    def _quantize(amount: Decimal, decimals: int) -> Decimal:
        """
        Quantizes the amount based on the token decimals.

        Parameters
        ----------
        amount: Decimal
            The amount.
        decimals: int
            The token decimals.

        Returns
        -------
        Decimal
            The quantized amount.
        """
        # print(f"[_quantize], amount: {amount}, type = {type(amount)}, decimals: {decimals}, type {type(decimals)}")

        if "." not in str(amount):
            return Decimal(str(amount))
        amount = format(amount, 'f')
        amount_num = str(amount).split(".")[0]
        amount_dec = str(amount).split(".")[1]
        # print(f"[_quantize], amount_dec: {amount_dec}, type = {type(amount_dec)}")
        amount_dec = str(amount_dec)[:int(decimals)]
        # print(f"[_quantize], amount_dec: {amount_dec}, type = {type(amount_dec)}")
        try:
            return Decimal(f"{str(amount_num)}.{amount_dec}")
        except Exception as e:
            #print("Error quantizing amount: ", f"{str(amount_num)}.{amount_dec}")
            pass



    def _get_token_address(self, token_key: str) -> str:
        """
        Gets the token address based on the token key.

        Parameters
        ----------
        token_key: str
            The token key (e.g. 'DAI-1d46')

        Returns
        -------
        str
            The token address.
        """
        return self._get_token(token_key).address

    def _get_token_decimals(self, token_key: str) -> int:
        """
        Gets the token decimals based on the token key.

        Parameters
        ----------
        token_key: str
            The token key (e.g. 'DAI-1d46')

        Returns
        -------
        int
            The token decimals.
        """
        return self._get_token(token_key).decimals

    def _get_token(self, token_key: str) -> Token:
        """
        Gets the token object based on the token key.

        Parameters
        ----------
        token_key: str
            The token key (e.g. 'DAI-1d46')

        Returns
        -------
        Token
            The token object.
        """

        return self.db.get_token(key=token_key)

    def _get_pool(self) -> Pool:
        """
        Gets the pool object based on the pool cid.

        Returns
        -------
        Pool
            The pool object.
        """
        return self.db.get_pool(cid=self.cid)

    @staticmethod
    def _convert_to_wei(amount: Union[int, Decimal, float], decimals: int) -> int:
        """
        Converts the amount to wei.

        Parameters
        ----------
        amount: int, Decimal or float
            The amount.
        decimals: int
            The token decimals.

        Returns
        -------
        int
            The amount in wei.
        """
        return int(amount * 10**decimals)

    @staticmethod
    def _convert_to_decimals(
        amount: Union[int, Decimal, float], decimals: int
    ) -> Decimal:
        """
        Converts the amount to decimals.

        Parameters
        ----------
        amount: int, Decimal or float
            The amount.
        decimals: int
            The token decimals.

        Returns
        -------
        Decimal
            The amount in decimals.
        """
        return Decimal(amount) / Decimal(10**decimals)

    def _check_if_carbon(self) -> bool:
        """
        Checks if the curve is a Carbon curve.

        Returns
        -------
        bool
            Whether the curve is a Carbon curve.
        """
        if "-" in self.cid:
            self._cid_tkn = self.cid.split("-")[1]
            self.cid = self.cid.split("-")[0]
            return True
        return False

    @property
    def cid_tkn(self) -> str:
        """
        The token cid.
        """
        return self._cid_tkn

    @property
    def tknin_address(self) -> str:
        """
        The input token address.
        """
        return self._tknin_address

    @property
    def tknin_decimals(self) -> int:
        """
        The input token decimals.
        """
        return self._tknin_decimals

    @property
    def amtin_wei(self) -> int:
        """
        The input amount in wei.
        """
        return self._amtin_wei

    @property
    def amtin_decimals(self) -> Decimal:
        """
        The input amount in decimals.
        """
        return self._amtin_decimals

    @property
    def tknout_address(self) -> str:
        """
        The output token address.
        """
        return self._tknout_address

    @property
    def tknout_decimals(self) -> int:
        """
        The output token decimals.
        """
        return self._tknout_decimals

    @property
    def amtout_wei(self) -> int:
        """
        The output amount in wei.
        """
        return self._amtout_wei

    @property
    def amtout_decimals(self) -> Decimal:
        """
        The output amount in decimals.
        """
        return self._amtout_decimals

    @property
    def is_carbon(self) -> bool:
        """
        Whether the curve is a Carbon curve.
        """
        return self._is_carbon

    @property
    def pool(self) -> Pool:
        """
        The pool object.
        """
        return self._get_pool()

    @property
    def token_in(self) -> Token:
        """
        The input token object.
        """
        return self._get_token(self.tknin)

    @property
    def token_out(self) -> Token:
        """
        The output token object.
        """
        return self._get_token(self.tknout)

    @property
    def amtin_quantized(self) -> Decimal:
        """
        Returns
        -------
        Decimal
            The quantized input amount.
        """
        return self._amtin_quantized

    @property
    def amtout_quantized(self) -> Decimal:
        """
        Returns
        -------
        Decimal
            The quantized output amount.
        """
        return self._amtout_quantized



# coding=utf-8
"""
Contains the base class for pools. This class is responsible for handling pools and updating the state of the pools.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, Any

from web3.contract import Contract


@dataclass
class Pool(ABC):
    """
    Abstract base class representing a pool.

    Attributes
    ----------
    state : Dict[str, Any]
        The pool state.
    """

    __VERSION__ = "0.0.1"
    __DATE__ = "2023-07-03"

    state: Dict[str, Any] = field(default_factory=dict)

    @staticmethod
    def get_common_data(
        event: Dict[str, Any], pool_info: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Get common (common to all Pool child classes) data from an event and pool info.

        Args:
            event (Dict[str, Any]): The event data.
            pool_info (Dict[str, Any]): The pool information.

        Returns:
            Dict[str, Any]: A dictionary containing common data extracted from the event and pool info.
        """
        return {
            "last_updated_block": event["blockNumber"],
            "timestamp": datetime.now().strftime("%H:%M:%S"),
            "pair_name": pool_info["pair_name"],
            "descr": pool_info["descr"],
            "address": event["address"],
        }

    @staticmethod
    @abstractmethod
    def update_from_event(
        event_args: Dict[str, Any], data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Update the pool state from an event.

        Parameters
        ----------
        event_args : Dict[str, Any]
            The event arguments.
        data : Dict[str, Any]
            The pool data.

        Returns
        -------
        Dict[str, Any]
            The updated pool data.
        """
        pass

    @abstractmethod
    def update_from_contract(self, contract: Contract) -> Dict[str, Any]:
        """
        Update the pool state from a contract.

        Parameters
        ----------
        contract : Contract
            The contract.

        Returns
        -------
        Dict[str, Any]
            The updated pool data.
        """
        pass

    @staticmethod
    @abstractmethod
    def unique_key() -> str:
        """
        Returns the unique key for the pool.
        """
        pass



"""
Utility functions for FastLane project

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
import datetime
import math
import os.path
from _decimal import Decimal
from dataclasses import dataclass
from typing import Tuple, List, Dict, Any

import pandas as pd
import requests
from web3 import Web3
from web3.contract import Contract

from fastlane_bot.config import config as cfg
from fastlane_bot.data.abi import *


def convert_decimals_to_wei_format(tkn_amt: Decimal, decimals: int) -> int:
    """
    param: tkn_amt: the number of tokens to convert
    param: token: the name of the token

    Returns:
    The number of tokens in WEI format according to the decimals used by the token
    """
    decimals = Decimal(str(decimals))
    tkn_amt = Decimal(str(tkn_amt))
    if decimals == 0:
        decimals = Decimal("1")
    return int(Decimal(tkn_amt * 10**decimals))

def num_format(number):
    try:
        return '{0:.4f}'.format(number)
    except Exception as e:
        return number

def num_format_float(number):
    try:
        return float('{0:.4f}'.format(number))
    except Exception as e:
        return number

def log_format(log_data: {}, log_name: str = "new"):
    now = datetime.datetime.now()
    time_ts = str(int(now.timestamp()))  # timestamp (epoch)
    time_iso = now.isoformat().split('.')[0]
    # print(time_ts)
    # print(time_iso)

    log_string = f"[{time_iso}::{time_ts}] |{log_name}| == {log_data}"
    return log_string
    # return "\n".join("[{" + time_iso + "}::{" + time_ts + "}] |" + log_name + "| == {d}\n".format(d=(log_data)))


def get_coingecko_token_table() -> List[Dict[str, Any]]:
    """
    Get the token table from coingecko
    :return:  list of tokens
    """
    token_list = requests.get(url=cfg.COINGECKO_URL).json()["tokens"]

    tokens = [
        {
            "address": cfg.w3.toChecksumAddress(token["address"]),
            "symbol": token["symbol"],
            "decimals": token["decimals"],
            "name": token["name"],
        }
        for token in token_list
    ]
    tokens.append(
        {"address": cfg.ETH_ADDRESS, "symbol": "ETH", "decimals": 18, "name": "ETH"}
    )

    return tokens


# Initialize Contracts
def initialize_contract_with_abi(address: str, abi: List[Any], web3: Web3) -> Contract:
    """
    Initialize a contract with an abi
    :param address:  address of the contract
    :param abi:  abi of the contract
    :param web3:  web3 instance
    :return:  contract instance
    """
    return web3.eth.contract(address=address, abi=abi)


def initialize_contract_without_abi(address: str, web3):
    abi_endpoint = f"https://api.etherscan.io/api?module=contract&action=getabi&address={address}&apikey={cfg.ETHERSCAN_TOKEN}"
    abi = json.loads(requests.get(abi_endpoint).text)
    return web3.eth.contract(address=address, abi=abi["result"])


def initialize_contract(web3, address: str, abi=None) -> Contract:
    """
    Initialize a contract with an abi
    :param web3:    web3 instance
    :param address:  address of the contract
    :param abi:  abi of the contract
    :return:  contract instance
    """
    if abi is None:
        return initialize_contract_without_abi(address=address, web3=web3)
    else:
        return initialize_contract_with_abi(address=address, abi=abi, web3=web3)


def get_contract_from_abi(address: str, exchange_name: str) -> Contract:
    """
    The contract of the exchange.
    """
    return Contract.from_abi(
        name=f"{address}",
        address=f"{address}",
        abi=get_abi_and_router(exchange_name)[0],
    )


def convert_decimals(amt: Decimal, n: int) -> Decimal:
    """
    Utility function to convert to Decimaling point value of a specific precision.
    """
    if amt is None:
        return Decimal("0")
    return Decimal(str(amt / (Decimal("10") ** Decimal(str(n)))))


def get_abi_and_router(exchange: str) -> Tuple[list, str]:
    """
    Returns the ABI and router address for the pool

    :param exchange: The exchange to get the ABI and POOL_INFO_FOR_EXCHANGE for

    :return: The ABI and POOL_INFO_FOR_EXCHANGE
    """
    base_path = os.path.normpath(os.path.dirname(os.path.abspath(__file__)))
    POOL_INFO_FOR_EXCHANGE = pd.read_csv(f"{base_path}/pairs.csv")
    if exchange == cfg.BANCOR_V2_NAME:
        POOL_INFO_FOR_EXCHANGE = POOL_INFO_FOR_EXCHANGE[
            POOL_INFO_FOR_EXCHANGE["exchange"] == cfg.BANCOR_V2_NAME
        ]
        ABI = BANCOR_V2_CONVERTER_ABI
    elif exchange == cfg.BANCOR_V3_NAME:
        POOL_INFO_FOR_EXCHANGE = POOL_INFO_FOR_EXCHANGE[
            POOL_INFO_FOR_EXCHANGE["exchange"] == cfg.BANCOR_V3_NAME
        ]
        ABI = BANCOR_V3_NETWORK_INFO_ABI
    elif exchange == cfg.SUSHISWAP_V2_NAME:
        POOL_INFO_FOR_EXCHANGE = POOL_INFO_FOR_EXCHANGE[
            POOL_INFO_FOR_EXCHANGE["exchange"] == cfg.SUSHISWAP_V2_NAME
        ]
        ABI = SUSHISWAP_POOLS_ABI
    elif exchange == cfg.UNISWAP_V2_NAME:
        POOL_INFO_FOR_EXCHANGE = POOL_INFO_FOR_EXCHANGE[
            POOL_INFO_FOR_EXCHANGE["exchange"] == cfg.UNISWAP_V2_NAME
        ]
        ABI = UNISWAP_V2_POOL_ABI
    elif exchange == cfg.UNISWAP_V3_NAME:
        ABI = UNISWAP_V3_POOL_ABI
        POOL_INFO_FOR_EXCHANGE = POOL_INFO_FOR_EXCHANGE[
            POOL_INFO_FOR_EXCHANGE["exchange"] == cfg.UNISWAP_V3_NAME
        ]
    return ABI, POOL_INFO_FOR_EXCHANGE


class HexbytesEncoder(json.JSONEncoder):
    def default(self, obj):
        return obj.hex() if isinstance(obj, Hexbytes) else super().default(obj)


class Hexbytes:
    def __init__(self, value):
        self.value = value

    def hex(self):
        return self.value.hex()


@dataclass
class EncodedOrder:
    """
    a single curve as encoded by the SDK

    :token:      token address
    :y:          number of token wei to sell on the curve
    :z:          curve capacity in number of token wei
    :A:          curve parameter A, multiplied by 2**48, encoded
    :B:          curve parameter B, multiplied by 2**48, encoded
    ----
    :A_:         curve parameter A in proper units
    :B_:         curve parameter B in proper units
    :p_start:    start token wei price of the order (in dy/dx)
    :p_end:      end token wei price of the order (in dy/dx)
    """

    token: str
    y: int
    z: int
    A: int
    B: int

    ONE = 2**48

    @classmethod
    def from_sdk(cls, token, order):
        return cls(token=token, **order)

    @property
    def descr(self):
        s = self
        return f"selling {s.token} @ ({1/s.p_start}..{1/s.p_end})  [TKNwei] per {s.token}wei"

    def __getitem__(self, item):
        return getattr(self, item)

    @classmethod
    def decodeFloat(cls, value):
        """undoes the mantisse/exponent encoding in A,B"""
        return (value % cls.ONE) << (value // cls.ONE)

    @classmethod
    def decode(cls, value):
        """decodes A,B to float"""
        return cls.decodeFloat(int(value)) / cls.ONE

    @staticmethod
    def bitLength(value):
        "minimal number of bits needed to represent the value"
        return len(bin(value).lstrip("0b")) if value > 0 else 0

    @classmethod
    def encodeRate(cls, value):
        "encodes a rate float to an A,B (using cls.ONE for scaling)"
        data = int(math.sqrt(value) * cls.ONE)
        length = cls.bitLength(data // cls.ONE)
        return (data >> length) << length

    @classmethod
    def encodeFloat(cls, value):
        "encodes a long int value as mantisse/exponent into a shorter integer"
        exponent = cls.bitLength(value // cls.ONE)
        mantissa = value >> exponent
        return mantissa | (exponent * cls.ONE)

    @classmethod
    def encode(cls, y, p_start_hi, p_end_lo, p_marg=None, z=None, token=None):
        """
        alternative constructor: creates a new encoded order from the given parameters

        :y:             number of token wei currently available to sell on the curve (loading)
        :p_start_hi:    start token wei price of the order (in dy/dx; highest)
        :p_end_lo:      end token wei price of the order (in dy/dx; lowest)
        :p_marg:        marginal token wei price of the order (in dy/dx)*
        :z:             curve capacity in number of token wei*

        *at most one of p_marg and z can be given; if neither is given,
        z is assumed to be equal to y
        """
        if token is None:
            token = "TKN"
        if p_marg is not None and z is not None:
            raise ValueError("at most one of p_marg and z can be given")
        if z is None:
            if p_marg is not None and p_marg >= p_start_hi or p_marg is None:
                z = y
            else:
                z = y * (p_start_hi - p_end_lo) // (p_marg - p_start_hi)
        return cls(
            y=int(y),
            z=int(z),
            A=cls.encodeFloat(cls.encodeRate(p_start_hi) - cls.encodeRate(p_end_lo)),
            B=cls.encodeFloat(cls.encodeRate(p_end_lo)),
            token=token,
        )

    @classmethod
    def encode_yzAB(cls, y, z, A, B, token=None):
        """
        encode A,B into the SDK format

        :y:    number of token wei currently available to sell on the curve (loading; as int)
        :z:    curve capacity in number of token wei (as int)
        :A:    curve parameter A (as float)
        :B:    curve parameter B (as float)
        """
        return cls(
            y=int(y),
            z=int(z),
            A=cls.encodeFloat(A),
            B=cls.encodeFloat(B),
            token=str(token),
        )

    @dataclass
    class DecodedOrder:
        """
        a single curve with the values of A,B decoded and as floats
        """

        y: int
        z: int
        A: float
        B: float

    @property
    def decoded(self):
        """
        returns a the order with A, B decoded as floats
        """
        return self.DecodedOrder(y=self.y, z=self.z, A=self.A_, B=self.B_)

    @property
    def A_(self):
        return self.decode(self.A)

    @property
    def B_(self):
        return self.decode(self.B)

    @property
    def p_end(self):
        return self.B_ * self.B_

    @property
    def p_start(self):
        return (self.B_ + self.A_) ** 2

    @property
    def p_marg(self):
        if self.y == self.z:
            return self.p_start
        elif self.y == 0:
            return self.p_end
        raise NotImplementedError("p_marg not implemented for non-full / empty orders")
        A = self.decodeFloat(self.A)
        B = self.decodeFloat(self.B)
        return self.decode(B + A * self.y / self.z) ** 2
        # https://github.com/bancorprotocol/carbon-simulator/blob/beta/benchmark/core/trade/impl.py
        # 'marginalRate' : decodeRate(B + A if y == z else B + A * y / z),


@dataclass
class UniV3Helper:
    """
    Handles math for Uni V3 pools.
    """

    contract_initialized: bool
    fee: str
    tick: int
    sqrt_price_q96: Decimal
    liquidity: Decimal

    _sqrt_price_q96_upper_bound: Decimal = None
    _sqrt_price_q96_lower_bound: Decimal = None
    TICK_BASE = Decimal("1.0001")
    Q96 = Decimal("2") ** Decimal("96")
    tkn0_decimal: int = None
    tkn1_decimal: int = None
    tick_spacing: int = None

    @staticmethod
    def dec_factor(dtkn0: int, dtkn1: int) -> Decimal:
        """
        Returns the decimal factor.
        """
        return Decimal(str(10 ** (dtkn0 - dtkn1)))

    def sqrt_price_x96_to_uint(
        self, sqrt_price_x96: Decimal, decimals_token0: int, decimals_token1: int
    ) -> Decimal:
        """
        Returns the sqrt price.
        """

        numerator1 = Decimal(str(sqrt_price_x96)) ** Decimal("2")
        numerator2 = self.dec_factor(decimals_token0, decimals_token1)
        return Decimal(str(numerator1 * numerator2 // Decimal(str(2**192))))

    @property
    def Pmarg(self) -> Decimal:
        """
        Returns the margin price.
        """
        return self.sqrt_price_x96_to_uint(
            self.sqrt_price_q96, self.tkn0_decimal, self.tkn1_decimal
        )

    @property
    def Pa(self) -> Decimal:
        """
        Returns the price of token 0.
        """
        return self.sqrt_price_x96_to_uint(
            self.sqrt_price_q96_lower_bound, self.tkn0_decimal, self.tkn1_decimal
        )

    @property
    def Pb(self) -> Decimal:
        """
        Returns the price of token 1.
        """
        return self.sqrt_price_x96_to_uint(
            self.sqrt_price_q96_upper_bound, self.tkn0_decimal, self.tkn1_decimal
        )

    @property
    def L(self) -> Decimal:
        """
        Returns the liquidity of the pool.
        """
        return self.liquidity / Decimal("10") ** (
            Decimal("0.5") * (self.tkn0_decimal + self.tkn1_decimal)
        )

    @property
    def sqrt_price_q96_upper_bound(self) -> Decimal:
        """
        Returns the upper bound of the price range.
        """
        return (
            self._sqrt_price_q96_upper_bound
            if self._sqrt_price_q96_upper_bound is not None
            else self.tick_to_sqrt_price_q96(self.upper_tick)
        )

    @property
    def sqrt_price_q96_lower_bound(self) -> Decimal:
        """
        Returns the lower bound of the price range.
        """
        return (
            self._sqrt_price_q96_lower_bound
            if self._sqrt_price_q96_lower_bound is not None
            else self.tick_to_sqrt_price_q96(self.lower_tick)
        )

    def tick_to_sqrt_price_q96(self, tick: Decimal) -> Decimal:
        """returns the price given a tick"""
        return (
            Decimal((self.TICK_BASE ** Decimal((tick / Decimal("2")))) * self.Q96)
            if self.contract_initialized
            else None
        )

    @property
    def lower_tick(self) -> Decimal:
        """
        Returns the lower tick of the pool.
        """
        if self.contract_initialized:
            return Decimal(
                math.floor(self.tick / self.tick_spacing) * self.tick_spacing
            )
        else:
            return Decimal(0)

    @property
    def upper_tick(self) -> Decimal:
        """
        Returns the upper tick of the pool.
        """
        if self.contract_initialized:
            return Decimal(self.lower_tick + self.tick_spacing)
        else:
            return Decimal(0)

    def __post_init__(self):
        self._sqrt_price_q96_upper_bound = self.tick_to_sqrt_price_q96(self.upper_tick)
        self._sqrt_price_q96_lower_bound = self.tick_to_sqrt_price_q96(self.lower_tick)



# coding=utf-8
"""
Triangle single arbitrage finder mode

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from typing import Union, List, Tuple, Any

from fastlane_bot.modes.base_triangle import ArbitrageFinderTriangleBase
from fastlane_bot.tools.cpc import CPCContainer
from fastlane_bot.tools.optimizer import CPCArbOptimizer


class ArbitrageFinderTriangleSingle(ArbitrageFinderTriangleBase):
    """
    Triangle single arbitrage finder mode
    """

    arb_mode = "single_triangle"

    def find_arbitrage(self, candidates: List[Any] = None, ops: Tuple = None, best_profit: float = 0, profit_src: float = 0) -> Union[List, Tuple]:
        """
        see base.py
        """

        if candidates is None:
            candidates = []

        # Get combinations of flashloan tokens
        combos = self.get_combos(
            self.flashloan_tokens, self.CCm, arb_mode=self.arb_mode
        )

        # Check each source token and miniverse combination
        for src_token, miniverse in combos:
            r = None

            # Instantiate the container and optimizer objects
            CC_cc = CPCContainer(miniverse)
            O = CPCArbOptimizer(CC_cc)

            try:
                # Perform the optimization
                r = O.margp_optimizer(src_token)

                # Get the profit in the source token
                profit_src = -r.result

                # Get trade instructions in different formats
                trade_instructions_df = r.trade_instructions(O.TIF_DFAGGR)
                trade_instructions_dic = r.trade_instructions(O.TIF_DICTS)
                trade_instructions = r.trade_instructions()
            except Exception:
                continue

            # Get the candidate ids
            cids = [ti["cid"] for ti in trade_instructions_dic]

            # Calculate the profit
            profit = self.calculate_profit(src_token, profit_src, self.CCm, cids)

            if str(profit) == "nan":
                self.ConfigObj.logger.debug("profit is nan, skipping")
                continue

            # Handle candidates based on conditions
            candidates += self.handle_candidates(
                best_profit,
                profit,
                trade_instructions_df,
                trade_instructions_dic,
                src_token,
                trade_instructions,
            )

            # Find the best operations
            best_profit, ops = self.find_best_operations(
                best_profit,
                ops,
                profit,
                trade_instructions_df,
                trade_instructions_dic,
                src_token,
                trade_instructions,
            )

        return candidates if self.result == self.AO_CANDIDATES else ops



# coding=utf-8
"""
Multi-pairwise arbitrage finder mode

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from typing import List, Any, Tuple, Union, Hashable

import pandas as pd

from fastlane_bot.modes.base_pairwise import ArbitrageFinderPairwiseBase
from fastlane_bot.tools.cpc import CPCContainer
from fastlane_bot.tools.optimizer import CPCArbOptimizer


class FindArbitrageMultiPairwise(ArbitrageFinderPairwiseBase):
    """
    Multi-pairwise arbitrage finder mode.
    """

    arb_mode = "multi_pairwise"

    def find_arbitrage(self, candidates: List[Any] = None, ops: Tuple = None, best_profit: float = 0, profit_src: float = 0) -> Union[List, Tuple]:
        """
        see base.py
        """
        if self.base_exchange != "carbon_v1":
            raise ValueError("base_exchange must be carbon_v1 for `multi` mode")

        if candidates is None:
            candidates = []

        all_tokens, combos = self.get_combos(self.CCm, self.flashloan_tokens)
        if self.result == self.AO_TOKENS:
            return all_tokens, combos

        candidates = []
        self.ConfigObj.logger.debug(
            f"\n ************ combos: {len(combos)} ************\n"
        )
        for tkn0, tkn1 in combos:
            r = None
            CC = self.CCm.bypairs(f"{tkn0}/{tkn1}")
            if len(CC) < 2:
                continue
            carbon_curves = [x for x in CC.curves if x.params.exchange == "carbon_v1"]
            not_carbon_curves = [
                x for x in CC.curves if x.params.exchange != "carbon_v1"
            ]
            curve_combos = [[curve] + carbon_curves for curve in not_carbon_curves]

            for curve_combo in curve_combos:
                src_token = tkn1

                if len(curve_combo) < 2:
                    continue

                try:
                    (
                        O,
                        profit_src,
                        r,
                        trade_instructions_df,
                    ) = self.run_main_flow(curves=curve_combo, src_token=src_token, tkn0=tkn0, tkn1=tkn1)

                    non_carbon_cids = [
                        curve.cid
                        for curve in curve_combo
                        if curve.params.get("exchange") != "carbon_v1"
                    ]
                    non_carbon_row = trade_instructions_df.loc[non_carbon_cids[0]]
                    tkn0_into_carbon = non_carbon_row[0] < 0
                    wrong_direction_cids = self.get_wrong_direction_cids(
                        tkn0_into_carbon, trade_instructions_df
                    )

                    if non_carbon_cids and len(wrong_direction_cids) > 0:
                        filtered_curves = self.process_wrong_direction_pools(
                            curve_combo=curve_combo, wrong_direction_cids=wrong_direction_cids
                        )
                        if len(filtered_curves) < 2:
                            continue
                        (
                            O,
                            profit_src,
                            r,
                            trade_instructions_df,
                        ) = self.run_main_flow(curves=filtered_curves, src_token=src_token, tkn0=tkn0, tkn1=tkn1)
                    trade_instructions_dic = r.trade_instructions(O.TIF_DICTS)
                    trade_instructions = r.trade_instructions()

                except Exception:
                    continue

                # Get the cids
                cids = [ti["cid"] for ti in trade_instructions_dic]

                # Calculate the profit
                profit = self.calculate_profit(src_token, profit_src, self.CCm, cids)

                if str(profit) == "nan":
                    self.ConfigObj.logger.debug("profit is nan, skipping")
                    continue

                # Handle candidates based on conditions
                candidates += self.handle_candidates(
                    best_profit,
                    profit,
                    trade_instructions_df,
                    trade_instructions_dic,
                    src_token,
                    trade_instructions,
                )

                # Find the best operations
                best_profit, ops = self.find_best_operations(
                    best_profit,
                    ops,
                    profit,
                    trade_instructions_df,
                    trade_instructions_dic,
                    src_token,
                    trade_instructions,
                )

        return candidates if self.result == self.AO_CANDIDATES else ops

    def get_wrong_direction_cids(
        self, tkn0_into_carbon: bool, trade_instructions_df: pd.DataFrame
    ) -> List[Hashable]:
        """
        Get the cids of the wrong direction curves

        Parameters
        ----------
        tkn0_into_carbon : bool
            True if tkn0 is being converted into carbon, False otherwise
        trade_instructions_df : pd.DataFrame
            The trade instructions dataframe

        Returns
        -------
        List[str]
            The cids of the wrong direction curves
        """
        return [
            idx
            for idx, row in trade_instructions_df.iterrows()
            if (
                (tkn0_into_carbon and row[0] < 0)
                or (not tkn0_into_carbon and row[0] > 0)
            )
            and ("-0" in idx or "-1" in idx)
        ]

    @staticmethod
    def run_main_flow(
        curves: List[Any], src_token: str, tkn0: str, tkn1: str
    ) -> Tuple[Any, float, Any, pd.DataFrame]:
        """
        Run main flow to find arbitrage.
        """
        CC_cc = CPCContainer(curves)
        O = CPCArbOptimizer(CC_cc)
        pstart = {
            tkn0: CC_cc.bypairs(f"{tkn0}/{tkn1}")[0].p
        }  # this intentionally selects the non_carbon curve
        r = O.margp_optimizer(src_token, params=dict(pstart=pstart))
        profit_src = -r.result
        trade_instructions_df = r.trade_instructions(O.TIF_DFAGGR)
        return O, profit_src, r, trade_instructions_df

    def process_wrong_direction_pools(
        self, curve_combo: List[Any], wrong_direction_cids: List[Hashable]
    ) -> [str]:
        """
        Process curves with wrong direction pools.
        """
        new_curves = [
            curve for curve in curve_combo if curve.cid not in wrong_direction_cids
        ]
        return new_curves
